//  ************************************************************************************************
//
//  BornAgain: simulate and fit reflection and scattering
//
//! @file      GUI/Model/Sample/ParticleLayoutItem.h
//! @brief     Defines class ParticleLayoutItem
//!
//! @homepage  http://www.bornagainproject.org
//! @license   GNU General Public License v3 or higher (see COPYING)
//! @copyright Forschungszentrum Jülich GmbH 2018
//! @authors   Scientific Computing Group at MLZ (see CITATION, AUTHORS)
//
//  ************************************************************************************************

#ifndef BORNAGAIN_GUI_MODEL_SAMPLE_PARTICLELAYOUTITEM_H
#define BORNAGAIN_GUI_MODEL_SAMPLE_PARTICLELAYOUTITEM_H

#include "GUI/Model/Group/SelectionDescriptor.h"
#include "GUI/Model/Session/SessionModel.h"

class InterferenceItem;
class ItemWithParticles;
class DoubleDescriptor;

class BA_CORE_API_ ParticleLayoutItem : public SessionItem {
private:
    static constexpr auto P_TOTAL_DENSITY{"TotalParticleDensity"};
    static constexpr auto P_WEIGHT{"Weight"};
    static constexpr auto T_PARTICLES{"Particle Tag"};
    static constexpr auto T_INTERFERENCE{"Interference Tag"};

public:
    static constexpr auto M_TYPE{"ParticleLayout"};

    ParticleLayoutItem();

    //! The density value which belonging only to the layout.
    //!
    //! This is the editable value. If an interference is present, this value may not be the one to
    //! be used for building the domain layout or to be presented. For the correct value-to-use,
    //! whether an interference is present or not, use totalDensity().
    DoubleDescriptor ownDensity() const;

    //! The real density.
    //!
    //! Returns ownDensity() if the interference is not influencing the density. If the
    //! interference is defining the density, this is the interference-calculated density.
    double totalDensity() const;

    DoubleDescriptor weight() const;

    //! The particles this layout contains.
    //!
    //! Note that this is not a hierarchical search! Only the topmost particles are returned, not
    //! their contained particles. For a hierarchical search use containedItemsWithParticles
    QVector<ItemWithParticles*> particles() const;
    void addParticle(ItemWithParticles* particle);
    void removeParticle(ItemWithParticles* particle);

    //! Return full hierarchical contained items with particles.
    //!
    //! \sa particles()
    QVector<ItemWithParticles*> containedItemsWithParticles() const;

    SelectionDescriptor<InterferenceItem*> interference() const;
    template <typename T> T* createInterference();
    void setInterference(InterferenceItem* interference);
    void removeInterference();

    // #baMigration Use only while not migrated from SessionModel!
    void enableDensity(bool b);

    //! Returns whether total density is defined by the currently selected interference.
    //!
    //! Two dimensional interference calculates density automatically; in these cases the "own"
    //! total density value should not be edited but set by the one from the interference.
    bool totalDensityIsDefinedByInterference() const;
};

template <typename T> T* ParticleLayoutItem::createInterference()
{
    static_assert(std::is_base_of<InterferenceItem, T>::value,
                  "Class must be derived from InterferenceFunciontItem");

    return model()->insertItem<T>(this, -1, T_INTERFERENCE);
}

#endif // BORNAGAIN_GUI_MODEL_SAMPLE_PARTICLELAYOUTITEM_H
